# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# compat imports
from __future__ import absolute_import, division, print_function
from builtins import (  # noqa
    bytes, dict, int, list, object, range, ascii, chr, hex, input,
    next, oct, open, pow, round, super, filter, map, zip)
# stdlib imports
import json
import logging
import multiprocessing
try:
    import pathlib2 as pathlib
except ImportError:  # noqa
    import pathlib
# non-stdlib imports
import click
import ruamel.yaml
# blobxfer library imports
import blobxfer.api
import blobxfer.util
# local imports
try:
    import cli.settings as settings
except (SystemError, ImportError):  # noqa
    try:
        from . import settings
    except (SystemError, ImportError):  # noqa
        # for local testing
        import settings

# create logger
logger = logging.getLogger('blobxfer')
# global defines
_CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


class CliContext(object):
    """CliContext class: holds context for CLI commands"""
    def __init__(self):
        """Ctor for CliContext"""
        self.config = None
        self.cli_options = {}
        self.credentials = None
        self.general_options = None
        self.show_config = False

    def initialize(self, action):
        # type: (CliContext, settings.TransferAction) -> None
        """Initialize context
        :param CliContext self: this
        :param settings.TransferAction action: transfer action
        """
        self._init_config()
        self.general_options = settings.create_general_options(
            self.config, action)
        self.credentials = settings.create_azure_storage_credentials(
            self.config, self.general_options)

    def _read_yaml_file(self, yaml_file):
        # type: (CliContext, pathlib.Path) -> None
        """Read a yaml file into self.config
        :param CliContext self: this
        :param pathlib.Path yaml_file: yaml file to load
        """
        with yaml_file.open('r') as f:
            if self.config is None:
                self.config = ruamel.yaml.load(
                    f, Loader=ruamel.yaml.RoundTripLoader)
            else:
                self.config = blobxfer.util.merge_dict(
                    ruamel.yaml.load(f, Loader=ruamel.yaml.RoundTripLoader),
                    self.config)

    def _init_config(self):
        # type: (CliContext) -> None
        """Initializes configuration of the context
        :param CliContext self: this
        """
        # load yaml config file into memory
        if blobxfer.util.is_not_empty(self.cli_options['yaml_config']):
            yaml_config = pathlib.Path(self.cli_options['yaml_config'])
            self._read_yaml_file(yaml_config)
        if self.config is None:
            self.config = {}
        # merge "global" cli options with config
        settings.merge_global_settings(self.config, self.cli_options)
        # set log file if specified
        logfile = self.config['options'].get('log_file', None)
        blobxfer.util.setup_logger(logger, logfile)
        # set azure storage logging level
        azstorage_logger = logging.getLogger('azure.storage')
        if self.config['options'].get('enable_azure_storage_logger', False):
            blobxfer.util.setup_logger(azstorage_logger, logfile)
            azstorage_logger.setLevel(logging.INFO)
        else:
            # disable azure storage logging: setting logger level to CRITICAL
            # effectively disables logging from azure storage
            azstorage_logger.setLevel(logging.CRITICAL)
        # set verbose logging
        if self.config['options'].get('verbose', False):
            blobxfer.util.set_verbose_logger_handlers()
        # output mixed config
        if self.show_config:
            logger.debug('config: \n{}'.format(
                json.dumps(self.config, indent=4)))
            logger.debug('cli config: \n{}'.format(
                json.dumps(
                    self.cli_options[self.cli_options['_action']],
                    indent=4, sort_keys=True)))
        del self.show_config


# create a pass decorator for shared context between commands
pass_cli_context = click.make_pass_decorator(CliContext, ensure=True)


def _config_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['yaml_config'] = value
        return value
    return click.option(
        '--config',
        expose_value=False,
        default=None,
        help='YAML configuration file',
        envvar='BLOBXFER_CONFIG_FILE',
        callback=callback)(f)


def _connect_timeout_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['connect_timeout'] = value
        return value
    return click.option(
        '--connect-timeout',
        expose_value=False,
        type=float,
        default=None,
        help='Timeout, in seconds, applied to connect operations',
        callback=callback)(f)


def _crypto_processes_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['crypto_processes'] = value
        return value
    return click.option(
        '--crypto-processes',
        expose_value=False,
        type=int,
        default=None,
        help='Concurrent crypto processes (download only)',
        callback=callback)(f)


def _disk_threads_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['disk_threads'] = value
        return value
    return click.option(
        '--disk-threads',
        expose_value=False,
        type=int,
        default=None,
        help='Concurrent disk threads',
        callback=callback)(f)


def _dry_run_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['dry_run'] = value
        return value
    return click.option(
        '--dry-run',
        expose_value=False,
        is_flag=True,
        default=False,
        help='Dry run [False]',
        callback=callback)(f)


def _enable_azure_storage_logger_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['enable_azure_storage_logger'] = value
        return value
    return click.option(
        '--enable-azure-storage-logger',
        expose_value=False,
        is_flag=True,
        default=False,
        help='Enable Azure Storage logger output [False]',
        callback=callback)(f)


def _log_file_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['log_file'] = value
        return value
    return click.option(
        '--log-file',
        expose_value=False,
        default=None,
        help='Log to file specified; this must be specified for progress '
        'bar to show',
        callback=callback)(f)


def _max_retries_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['max_retries'] = value
        return value
    return click.option(
        '--max-retries',
        expose_value=False,
        type=int,
        default=None,
        help='Maximum number of retries for a request; negative values are '
        'unlimited [10]',
        callback=callback)(f)


def _md5_processes_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['md5_processes'] = value
        return value
    return click.option(
        '--md5-processes',
        expose_value=False,
        type=int,
        default=None,
        help='Concurrent MD5 processes',
        callback=callback)(f)


def _progress_bar_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['progress_bar'] = value
        return value
    return click.option(
        '--progress-bar/--no-progress-bar',
        expose_value=False,
        default=None,
        help='Display progress bar instead of console logs; log file must '
        'be specified [True]',
        callback=callback)(f)


def _proxy_host_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['proxy_host'] = value
        return value
    return click.option(
        '--proxy-host',
        expose_value=False,
        default=None,
        help='Proxy host in the format of IP:Port',
        envvar='BLOBXFER_PROXY_HOST',
        callback=callback)(f)


def _proxy_password_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['proxy_password'] = value
        return value
    return click.option(
        '--proxy-password',
        expose_value=False,
        default=None,
        help='Proxy password',
        envvar='BLOBXFER_PROXY_PASSWORD',
        callback=callback)(f)


def _proxy_username_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['proxy_username'] = value
        return value
    return click.option(
        '--proxy-username',
        expose_value=False,
        default=None,
        help='Proxy username',
        envvar='BLOBXFER_PROXY_USERNAME',
        callback=callback)(f)


def _read_timeout_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['read_timeout'] = value
        return value
    return click.option(
        '--read-timeout',
        expose_value=False,
        type=float,
        default=None,
        help='Timeout, in seconds, applied to read operations',
        callback=callback)(f)


def _resume_file_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['resume_file'] = value
        return value
    return click.option(
        '--resume-file',
        expose_value=False,
        default=None,
        help='Save or use resume file specified',
        callback=callback)(f)


def _show_config_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.show_config = value
        return value
    return click.option(
        '--show-config',
        expose_value=False,
        is_flag=True,
        help='Show configuration',
        callback=callback)(f)


def _timeout_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['timeout'] = value
        return value
    return click.option(
        '--timeout',
        expose_value=False,
        type=float,
        default=None,
        help='[DEPRECATED] Timeout, in seconds, applied to both connect '
        'and read operations',
        callback=callback)(f)


def _transfer_threads_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['transfer_threads'] = value
        return value
    return click.option(
        '--transfer-threads',
        expose_value=False,
        type=int,
        default=None,
        help='Concurrent transfer threads',
        callback=callback)(f)


def _verbose_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['verbose'] = value
        return value
    return click.option(
        '-v', '--verbose',
        expose_value=False,
        is_flag=True,
        default=None,
        help='Verbose output',
        callback=callback)(f)


def _quiet_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['quiet'] = value
        return value
    return click.option(
        '-q', '--quiet',
        expose_value=False,
        is_flag=True,
        default=None,
        help='Quiet mode',
        callback=callback)(f)


def _local_resource_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['local_resource'] = value
        return value
    return click.option(
        '--local-path',
        expose_value=False,
        default=None,
        help='Local path; use - for stdin',
        callback=callback)(f)


def _storage_account_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['storage_account'] = value
        return value
    return click.option(
        '--storage-account',
        expose_value=False,
        default=None,
        help='Storage account name',
        envvar='BLOBXFER_STORAGE_ACCOUNT',
        callback=callback)(f)


def _remote_path_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['remote_path'] = value
        return value
    return click.option(
        '--remote-path',
        expose_value=False,
        default=None,
        help='Remote path on Azure Storage',
        callback=callback)(f)


def _storage_url_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['storage_url'] = value
        return value
    return click.option(
        '--storage-url',
        expose_value=False,
        default=None,
        help='Azure Storage URL as source or destination (can include SAS)',
        envvar='BLOBXFER_STORAGE_URL',
        callback=callback)(f)


def common_options(f):
    f = _verbose_option(f)
    f = _transfer_threads_option(f)
    f = _timeout_option(f)
    f = _show_config_option(f)
    f = _resume_file_option(f)
    f = _read_timeout_option(f)
    f = _quiet_option(f)
    f = _proxy_username_option(f)
    f = _proxy_password_option(f)
    f = _proxy_host_option(f)
    f = _progress_bar_option(f)
    f = _md5_processes_option(f)
    f = _max_retries_option(f)
    f = _log_file_option(f)
    f = _enable_azure_storage_logger_option(f)
    f = _dry_run_option(f)
    f = _disk_threads_option(f)
    f = _crypto_processes_option(f)
    f = _connect_timeout_option(f)
    f = _config_option(f)
    return f


def upload_download_options(f):
    f = _remote_path_option(f)
    f = _storage_url_option(f)
    f = _storage_account_option(f)
    f = _local_resource_option(f)
    return f


def _access_key_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['access_key'] = value
        return value
    return click.option(
        '--storage-account-key',
        expose_value=False,
        default=None,
        help='Storage account access key',
        envvar='BLOBXFER_STORAGE_ACCOUNT_KEY',
        callback=callback)(f)


def _access_tier_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['access_tier'] = value
        return value
    return click.option(
        '--access-tier',
        expose_value=False,
        default=None,
        help='Access tier to apply to target (block blob only)',
        callback=callback)(f)


def _chunk_size_bytes_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['chunk_size_bytes'] = value
        return value
    return click.option(
        '--chunk-size-bytes',
        expose_value=False,
        type=int,
        default=None,
        help='Block or chunk size in bytes; set to 0 for auto-select '
        'on upload [0]',
        callback=callback)(f)


def _delete_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['delete'] = value
        return value
    return click.option(
        '--delete',
        expose_value=False,
        is_flag=True,
        default=None,
        help='Delete extraneous files on target [False]',
        callback=callback)(f)


def _distribution_mode(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['distribution_mode'] = value
        return value
    return click.option(
        '--distribution-mode',
        expose_value=False,
        default=None,
        help='Vectored IO distribution mode: disabled, replica, '
        'stripe [disabled]',
        callback=callback)(f)


def _endpoint_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['endpoint'] = value
        return value
    return click.option(
        '--endpoint',
        expose_value=False,
        default=None,
        help='Azure Storage endpoint [core.windows.net]',
        callback=callback)(f)


def _exclude_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['exclude'] = value
        return value
    return click.option(
        '--exclude',
        expose_value=False,
        multiple=True,
        default=None,
        help='Exclude pattern',
        callback=callback)(f)


def _file_attributes(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['file_attributes'] = value
        return value
    return click.option(
        '--file-attributes/--no-file-attributes',
        expose_value=False,
        default=None,
        help='Store or restore file attributes [False]',
        callback=callback)(f)


def _file_md5_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['file_md5'] = value
        return value
    return click.option(
        '--file-md5/--no-file-md5',
        expose_value=False,
        default=None,
        help='Compute file MD5 [False]',
        callback=callback)(f)


def _include_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['include'] = value
        return value
    return click.option(
        '--include',
        expose_value=False,
        multiple=True,
        default=None,
        help='Include pattern',
        callback=callback)(f)


def _mode_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['mode'] = value
        return value
    return click.option(
        '--mode',
        expose_value=False,
        default=None,
        help='Transfer mode: auto, append, block, file, page [auto]',
        callback=callback)(f)


def _one_shot_bytes_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['one_shot_bytes'] = value
        return value
    return click.option(
        '--one-shot-bytes',
        expose_value=False,
        type=int,
        default=None,
        help='File sizes less than or equal to the specified byte threshold '
        'will be uploaded as one-shot for block blobs; the valid range that '
        'can be specified is 0 to 256MiB [0]',
        callback=callback)(f)


def _overwrite_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['overwrite'] = value
        return value
    return click.option(
        '--overwrite/--no-overwrite',
        expose_value=False,
        default=None,
        help='Overwrite destination if exists. For append blobs, '
        '--no-overwrite will append to any existing blob. [True]',
        callback=callback)(f)


def _recursive_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['recursive'] = value
        return value
    return click.option(
        '--recursive/--no-recursive',
        expose_value=False,
        default=None,
        help='Recursive [True]',
        callback=callback)(f)


def _rename_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['rename'] = value
        return value
    return click.option(
        '--rename',
        expose_value=False,
        is_flag=True,
        default=None,
        help='Rename to specified destination for a single object. '
        'Automatically enabled with stdin source. [False]',
        callback=callback)(f)


def _restore_file_lmt_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['restore_file_lmt'] = value
        return value
    return click.option(
        '--restore-file-lmt',
        expose_value=False,
        is_flag=True,
        default=None,
        help='Set last modified time equal to the last modified property '
        'in Azure Storage [false]',
        callback=callback)(f)


def _rsa_private_key_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['rsa_private_key'] = value
        return value
    return click.option(
        '--rsa-private-key',
        expose_value=False,
        default=None,
        help='RSA private key PEM file',
        envvar='BLOBXFER_RSA_PRIVATE_KEY',
        callback=callback)(f)


def _rsa_private_key_passphrase_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['rsa_private_key_passphrase'] = value
        return value
    return click.option(
        '--rsa-private-key-passphrase',
        expose_value=False,
        default=None,
        help='RSA private key passphrase',
        envvar='BLOBXFER_RSA_PRIVATE_KEY_PASSPHRASE',
        callback=callback)(f)


def _rsa_public_key_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['rsa_public_key'] = value
        return value
    return click.option(
        '--rsa-public-key',
        expose_value=False,
        default=None,
        help='RSA public key PEM file',
        envvar='BLOBXFER_RSA_PUBLIC_KEY',
        callback=callback)(f)


def _sas_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['sas'] = value
        return value
    return click.option(
        '--sas',
        expose_value=False,
        default=None,
        help='Shared access signature',
        envvar='BLOBXFER_SAS',
        callback=callback)(f)


def _skip_on_filesize_match_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['skip_on_filesize_match'] = value
        return value
    return click.option(
        '--skip-on-filesize-match',
        expose_value=False,
        is_flag=True,
        default=None,
        help='Skip on equivalent file size [False]',
        callback=callback)(f)


def _skip_on_lmt_ge_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['skip_on_lmt_ge'] = value
        return value
    return click.option(
        '--skip-on-lmt-ge',
        expose_value=False,
        is_flag=True,
        default=None,
        help='Skip on last modified time greater than or equal to [False]',
        callback=callback)(f)


def _skip_on_md5_match_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['skip_on_md5_match'] = value
        return value
    return click.option(
        '--skip-on-md5-match',
        expose_value=False,
        is_flag=True,
        default=None,
        help='Skip on MD5 match [False]',
        callback=callback)(f)


def _stdin_as_page_blob_size_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['stdin_as_page_blob_size'] = value
        return value
    return click.option(
        '--stdin-as-page-blob-size',
        expose_value=False,
        type=int,
        default=None,
        help='Size of page blob with input from stdin [0]',
        callback=callback)(f)


def _strip_components_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['strip_components'] = value
        return value
    return click.option(
        '--strip-components',
        expose_value=False,
        type=int,
        default=None,
        help='Strip leading file path components: local path for upload '
        'or remote path for download [0]',
        callback=callback)(f)


def _stripe_chunk_size_bytes_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['stripe_chunk_size_bytes'] = value
        return value
    return click.option(
        '--stripe-chunk-size-bytes',
        expose_value=False,
        type=int,
        default=None,
        help='Vectored IO stripe width in bytes [1073741824]',
        callback=callback)(f)


def _sync_copy_dest_access_key_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['sync_copy_dest_access_key'] = value
        return value
    return click.option(
        '--sync-copy-dest-storage-account-key',
        expose_value=False,
        default=None,
        help='Storage account access key for synccopy destination',
        envvar='BLOBXFER_SYNC_COPY_DEST_STORAGE_ACCOUNT_KEY',
        callback=callback)(f)


def _sync_copy_dest_mode_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['sync_copy_dest_mode'] = value
        return value
    return click.option(
        '--sync-copy-dest-mode',
        expose_value=False,
        default=None,
        help='Mode for synccopy destination',
        callback=callback)(f)


def _sync_copy_dest_remote_path_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['sync_copy_dest_remote_path'] = value
        return value
    return click.option(
        '--sync-copy-dest-remote-path',
        expose_value=False,
        default=None,
        help='Remote path on Azure Storage for synccopy destination',
        callback=callback)(f)


def _sync_copy_dest_sas_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['sync_copy_dest_sas'] = value
        return value
    return click.option(
        '--sync-copy-dest-sas',
        expose_value=False,
        default=None,
        help='Shared access signature for synccopy destination',
        envvar='BLOBXFER_SYNC_COPY_DEST_SAS',
        callback=callback)(f)


def _sync_copy_dest_storage_account_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['sync_copy_dest_storage_account'] = value
        return value
    return click.option(
        '--sync-copy-dest-storage-account',
        expose_value=False,
        default=None,
        help='Storage account name for synccopy destination',
        envvar='BLOBXFER_SYNC_COPY_DEST_STORAGE_ACCOUNT',
        callback=callback)(f)


def _sync_copy_dest_storage_url_option(f):
    def callback(ctx, param, value):
        clictx = ctx.ensure_object(CliContext)
        clictx.cli_options['sync_copy_dest_storage_url'] = value
        return value
    return click.option(
        '--sync-copy-dest-storage-url',
        expose_value=False,
        default=None,
        help='Azure Storage URL as destination (can include SAS)',
        envvar='BLOBXFER_SYNC_COPY_DEST_STORAGE_URL',
        callback=callback)(f)


def upload_options(f):
    f = _stripe_chunk_size_bytes_option(f)
    f = _strip_components_option(f)
    f = _stdin_as_page_blob_size_option(f)
    f = _skip_on_md5_match_option(f)
    f = _skip_on_lmt_ge_option(f)
    f = _skip_on_filesize_match_option(f)
    f = _sas_option(f)
    f = _rsa_public_key_option(f)
    f = _rsa_private_key_passphrase_option(f)
    f = _rsa_private_key_option(f)
    f = _rename_option(f)
    f = _recursive_option(f)
    f = _overwrite_option(f)
    f = _one_shot_bytes_option(f)
    f = _mode_option(f)
    f = _include_option(f)
    f = _file_md5_option(f)
    f = _file_attributes(f)
    f = _exclude_option(f)
    f = _endpoint_option(f)
    f = _distribution_mode(f)
    f = _delete_option(f)
    f = _chunk_size_bytes_option(f)
    f = _access_tier_option(f)
    f = _access_key_option(f)
    return f


def download_options(f):
    f = _strip_components_option(f)
    f = _skip_on_md5_match_option(f)
    f = _skip_on_lmt_ge_option(f)
    f = _skip_on_filesize_match_option(f)
    f = _sas_option(f)
    f = _rsa_private_key_passphrase_option(f)
    f = _rsa_private_key_option(f)
    f = _restore_file_lmt_option(f)
    f = _rename_option(f)
    f = _recursive_option(f)
    f = _overwrite_option(f)
    f = _mode_option(f)
    f = _include_option(f)
    f = _file_md5_option(f)
    f = _file_attributes(f)
    f = _exclude_option(f)
    f = _endpoint_option(f)
    f = _delete_option(f)
    f = _chunk_size_bytes_option(f)
    f = _access_key_option(f)
    return f


def sync_copy_options(f):
    f = _sync_copy_dest_storage_url_option(f)
    f = _sync_copy_dest_storage_account_option(f)
    f = _sync_copy_dest_sas_option(f)
    f = _sync_copy_dest_remote_path_option(f)
    f = _sync_copy_dest_mode_option(f)
    f = _sync_copy_dest_access_key_option(f)
    f = _storage_account_option(f)
    f = _skip_on_md5_match_option(f)
    f = _skip_on_lmt_ge_option(f)
    f = _skip_on_filesize_match_option(f)
    f = _sas_option(f)
    f = _remote_path_option(f)
    f = _overwrite_option(f)
    f = _mode_option(f)
    f = _include_option(f)
    f = _exclude_option(f)
    f = _endpoint_option(f)
    f = _chunk_size_bytes_option(f)
    f = _access_tier_option(f)
    f = _access_key_option(f)
    return f


@click.group(context_settings=_CONTEXT_SETTINGS)
@click.version_option(version=blobxfer.__version__)
@click.pass_context
def cli(ctx):
    """blobxfer: Azure Storage transfer tool"""
    pass


@cli.command('download')
@upload_download_options
@download_options
@common_options
@pass_cli_context
def download(ctx):
    """Download blobs or files from Azure Storage"""
    settings.add_cli_options(ctx.cli_options, settings.TransferAction.Download)
    ctx.initialize(settings.TransferAction.Download)
    specs = settings.create_download_specifications(
        ctx.cli_options, ctx.config)
    del ctx.cli_options
    for spec in specs:
        blobxfer.api.Downloader(
            ctx.general_options, ctx.credentials, spec
        ).start()


@cli.command('synccopy')
@sync_copy_options
@common_options
@pass_cli_context
def synccopy(ctx):
    """Synchronously copy blobs or files between Azure Storage accounts"""
    settings.add_cli_options(ctx.cli_options, settings.TransferAction.Synccopy)
    ctx.initialize(settings.TransferAction.Synccopy)
    specs = settings.create_synccopy_specifications(
        ctx.cli_options, ctx.config)
    del ctx.cli_options
    for spec in specs:
        blobxfer.api.SyncCopy(
            ctx.general_options, ctx.credentials, spec
        ).start()


@cli.command('upload')
@upload_download_options
@upload_options
@common_options
@pass_cli_context
def upload(ctx):
    """Upload files to Azure Storage"""
    settings.add_cli_options(ctx.cli_options, settings.TransferAction.Upload)
    ctx.initialize(settings.TransferAction.Upload)
    specs = settings.create_upload_specifications(
        ctx.cli_options, ctx.config)
    del ctx.cli_options
    for spec in specs:
        blobxfer.api.Uploader(
            ctx.general_options, ctx.credentials, spec
        ).start()


if __name__ == '__main__':
    multiprocessing.freeze_support()
    cli()



# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# compat imports
from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
from builtins import (  # noqa
    bytes, dict, int, list, object, range, ascii, chr, hex, input,
    next, oct, open, pow, round, super, filter, map, zip)
# stdlib imports
# non-stdlib imports
# local imports


class Download(object):
    """Download resume object"""
    def __init__(
            self, final_path, length, chunk_size, next_integrity_chunk,
            completed, md5):
        # type: (Download, str, int, int, int, bool, str) -> None
        """Ctor for Download
        :param Download self: this
        :param str final_path: final path
        :param int length: total bytes
        :param int chunk_size: chunk size in bytes
        :param int next_integrity_chunk: next integrity chunk
        :param bool completed: completed
        :param str md5: md5 hex digest
        """
        self._final_path = final_path
        self._length = length
        self._chunk_size = chunk_size
        self._next_integrity_chunk = next_integrity_chunk
        self._completed = completed
        self._md5hexdigest = md5 if md5 is not None else None

    @property
    def final_path(self):
        # type: (Download) -> str
        """Final path
        :param Download self: this
        :rtype: str
        :return: final path
        """
        return self._final_path

    @property
    def length(self):
        # type: (Download) -> int
        """Content length
        :param Download self: this
        :rtype: int
        :return: number of bytes
        """
        return self._length

    @property
    def chunk_size(self):
        # type: (Download) -> int
        """Chunk size
        :param Download self: this
        :rtype: int
        :return: chunk size in bytes
        """
        return self._chunk_size

    @property
    def next_integrity_chunk(self):
        # type: (Download) -> int
        """Get Next integrity chunk
        :param Download self: this
        :rtype: int
        :return: next integrity chunk
        """
        return self._next_integrity_chunk

    @next_integrity_chunk.setter
    def next_integrity_chunk(self, value):
        # type: (Download) -> None
        """Set Next integrity chunk
        :param Download self: this
        :param int value: next chunk num
        """
        self._next_integrity_chunk = value

    @property
    def completed(self):
        # type: (Download) -> bool
        """Get Completed
        :param Download self: this
        :rtype: bool
        :return: if completed
        """
        return self._completed

    @completed.setter
    def completed(self, value):
        # type: (Download) -> None
        """Set Completed
        :param Download self: this
        :param bool value: completion value
        """
        self._completed = value

    @property
    def md5hexdigest(self):
        # type: (Download) -> str
        """Get md5 hex digest
        :param Download self: this
        :rtype: str
        :return: md5 hex digest
        """
        return self._md5hexdigest

    @md5hexdigest.setter
    def md5hexdigest(self, value):
        # type: (Download) -> None
        """Set md5 hex digest value if value is not None
        :param Download self: this
        :param str value: md5 hex digest
        """
        if value is None:
            return
        self._md5hexdigest = value

    def __repr__(self):
        # type: (Download) -> str
        """Return representation
        :param Download self: this
        :rtype: str
        :return: representation string
        """
        return ('Download<final_path={} length={} chunk_size={} '
                'next_integrity_chunk={} completed={} md5={}>').format(
                    self.final_path, self.length, self.chunk_size,
                    self.next_integrity_chunk, self.completed,
                    self.md5hexdigest)


class Upload(object):
    """Upload resume object"""
    def __init__(
            self, local_path, length, chunk_size, total_chunks,
            completed_chunks, completed, md5):
        # type: (Upload, str, int, int, int, int, bool, str) -> None
        """Ctor for Upload
        :param Upload self: this
        :param str local_path: local path
        :param int length: total bytes
        :param int chunk_size: chunk size in bytes
        :param int total_chunks: total chunks
        :param int completed_chunks: completed chunks
        :param bool completed: completed
        :param str md5: md5 hex digest
        """
        self._local_path = local_path
        self._length = length
        self._chunk_size = chunk_size
        self._total_chunks = total_chunks
        self._completed_chunks = completed_chunks
        self._completed = completed
        self._md5hexdigest = md5 if md5 is not None else None

    @property
    def local_path(self):
        # type: (Upload) -> str
        """Local path
        :param Upload self: this
        :rtype: str
        :return: local path
        """
        return self._local_path

    @property
    def length(self):
        # type: (Upload) -> int
        """Content length
        :param Upload self: this
        :rtype: int
        :return: number of bytes
        """
        return self._length

    @property
    def chunk_size(self):
        # type: (Upload) -> int
        """Chunk size
        :param Upload self: this
        :rtype: int
        :return: chunk size in bytes
        """
        return self._chunk_size

    @property
    def total_chunks(self):
        # type: (Upload) -> int
        """Get total number of chunks
        :param Upload self: this
        :rtype: int
        :return: total chunks
        """
        return self._total_chunks

    @property
    def completed_chunks(self):
        # type: (Upload) -> int
        """Get Completed chunks
        :param Upload self: this
        :rtype: int
        :return: completed chunks
        """
        return self._completed_chunks

    @completed_chunks.setter
    def completed_chunks(self, value):
        # type: (Upload, int) -> None
        """Set Completed chunks
        :param Upload self: this
        :param int value: completed chunks
        """
        self._completed_chunks = value

    @property
    def completed(self):
        # type: (Upload) -> bool
        """Get Completed
        :param Upload self: this
        :rtype: bool
        :return: if completed
        """
        return self._completed

    @completed.setter
    def completed(self, value):
        # type: (Upload) -> None
        """Set Completed
        :param Upload self: this
        :param bool value: completion value
        """
        self._completed = value

    @property
    def md5hexdigest(self):
        # type: (Upload) -> str
        """Get md5 hex digest
        :param Upload self: this
        :rtype: str
        :return: md5 hex digest
        """
        return self._md5hexdigest

    @md5hexdigest.setter
    def md5hexdigest(self, value):
        # type: (Upload) -> None
        """Set md5 hex digest value if value is not None
        :param Upload self: this
        :param str value: md5 hex digest
        """
        if value is None:
            return
        self._md5hexdigest = value

    def __repr__(self):
        # type: (Upload) -> str
        """Return representation
        :param Upload self: this
        :rtype: str
        :return: representation string
        """
        return ('Upload<local_path={} length={} chunk_size={} '
                'total_chunks={} completed_chunks={} completed={} '
                'md5={}>').format(
                    self.local_path, self.length, self.chunk_size,
                    self.total_chunks, self.completed_chunks, self.completed,
                    self.md5hexdigest)


class SyncCopy(object):
    """SyncCopy resume object"""
    def __init__(
            self, length, src_block_list, offset, chunk_size, total_chunks,
            completed_chunks, completed):
        # type: (SyncCopy, int, int, int, int, bool) -> None
        """Ctor for SyncCopy
        :param SyncCopy self: this
        :param int length: total bytes
        :param int chunk_size: chunk size in bytes
        :param int total_chunks: total chunks
        :param int completed_chunks: completed chunks
        :param bool completed: completed
        """
        self._length = length
        self._src_block_list = src_block_list
        self._offset = offset
        self._chunk_size = chunk_size
        self._total_chunks = total_chunks
        self._completed_chunks = completed_chunks
        self._completed = completed

    @property
    def length(self):
        # type: (SyncCopy) -> int
        """Content length
        :param SyncCopy self: this
        :rtype: int
        :return: number of bytes
        """
        return self._length

    @property
    def src_block_list(self):
        # type: (SyncCopy) -> list
        """Source committed block list
        :param SyncCopy self: this
        :rtype: list
        :return: source committed block list
        """
        return self._src_block_list

    @property
    def offset(self):
        # type: (SyncCopy) -> int
        """Current offset
        :param SyncCopy self: this
        :rtype: int
        :return: current offset
        """
        return self._offset

    @offset.setter
    def offset(self, value):
        # type: (SyncCopy, int) -> None
        """Set offset
        :param SyncCopy self: this
        :param int value: offset
        """
        self._offset = value

    @property
    def chunk_size(self):
        # type: (SyncCopy) -> int
        """Chunk size
        :param SyncCopy self: this
        :rtype: int
        :return: chunk size in bytes
        """
        return self._chunk_size

    @property
    def total_chunks(self):
        # type: (SyncCopy) -> int
        """Get total number of chunks
        :param SyncCopy self: this
        :rtype: int
        :return: total chunks
        """
        return self._total_chunks

    @property
    def completed_chunks(self):
        # type: (SyncCopy) -> int
        """Get Completed chunks
        :param SyncCopy self: this
        :rtype: int
        :return: completed chunks
        """
        return self._completed_chunks

    @completed_chunks.setter
    def completed_chunks(self, value):
        # type: (SyncCopy, int) -> None
        """Set Completed chunks
        :param SyncCopy self: this
        :param int value: completed chunks
        """
        self._completed_chunks = value

    @property
    def completed(self):
        # type: (SyncCopy) -> bool
        """Get Completed
        :param SyncCopy self: this
        :rtype: bool
        :return: if completed
        """
        return self._completed

    @completed.setter
    def completed(self, value):
        # type: (SyncCopy) -> None
        """Set Completed
        :param SyncCopy self: this
        :param bool value: completion value
        """
        self._completed = value

    def __repr__(self):
        # type: (SyncCopy) -> str
        """Return representation
        :param SyncCopy self: this
        :rtype: str
        :return: representation string
        """
        return ('SyncCopy<length={} chunk_size={} total_chunks={} '
                'completed_chunks={} completed={}>').format(
                    self.length, self.chunk_size, self.total_chunks,
                    self.completed_chunks, self.completed)



# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# compat imports
from __future__ import absolute_import, division, print_function
from builtins import (  # noqa
    bytes, dict, int, list, object, range, ascii, chr, hex, input,
    next, oct, open, pow, round, super, filter, map, zip
)
# stdlib imports
import logging
# non-stdlib imports
import azure.common
import azure.storage.blob.models
# local imports
import blobxfer.models.azure
import blobxfer.util

# create logger
logger = logging.getLogger(__name__)


def check_if_single_blob(client, container, prefix, timeout=None):
    # type: (azure.storage.blob.BaseBlobService, str, str, int) -> bool
    """Check if prefix is a single blob or multiple blobs
    :param azure.storage.blob.BaseBlobService client: blob client
    :param str container: container
    :param str prefix: path prefix
    :param int timeout: timeout
    :rtype: bool
    :return: if prefix in container is a single blob
    """
    if blobxfer.util.blob_is_snapshot(prefix):
        return True
    try:
        client.get_blob_properties(
            container_name=container, blob_name=prefix, timeout=timeout)
    except azure.common.AzureMissingResourceHttpError:
        return False
    return True


def get_blob_properties(client, container, prefix, mode, timeout=None):
    # type: (azure.storage.blob.BaseBlobService, str, str,
    #        blobxfer.models.azure.StorageModes, int) ->
    #        azure.storage.blob.models.Blob
    """Get blob properties
    :param azure.storage.blob.BaseBlobService client: blob client
    :param str container: container
    :param str prefix: path prefix
    :param blobxfer.models.azure.StorageModes mode: storage mode
    :param int timeout: timeout
    :rtype: azure.storage.blob.models.Blob
    :return: blob
    """
    if mode == blobxfer.models.azure.StorageModes.File:
        raise RuntimeError(
            'cannot list Azure Blobs with incompatible mode: {}'.format(
                mode))
    try:
        blob = client.get_blob_properties(
            container_name=container, blob_name=prefix, timeout=timeout)
    except azure.common.AzureMissingResourceHttpError:
        return None
    if ((mode == blobxfer.models.azure.StorageModes.Append and
         blob.properties.blob_type !=
         azure.storage.blob.models._BlobTypes.AppendBlob) or
            (mode == blobxfer.models.azure.StorageModes.Block and
             blob.properties.blob_type !=
             azure.storage.blob.models._BlobTypes.BlockBlob) or
            (mode == blobxfer.models.azure.StorageModes.Page and
             blob.properties.blob_type !=
             azure.storage.blob.models._BlobTypes.PageBlob)):
        raise RuntimeError(
            'existing blob type {} mismatch with mode {}'.format(
                blob.properties.blob_type, mode))
    return blob


def list_blobs(client, container, prefix, mode, recursive, timeout=None):
    # type: (azure.storage.blob.BaseBlobService, str, str,
    #        blobxfer.models.azure.StorageModes, bool, int) ->
    #        azure.storage.blob.models.Blob
    """List blobs in path conforming to mode
    :param azure.storage.blob.BaseBlobService client: blob client
    :param str container: container
    :param str prefix: path prefix
    :param blobxfer.models.azure.StorageModes mode: storage mode
    :param bool recursive: recursive
    :param int timeout: timeout
    :rtype: azure.storage.blob.models.Blob
    :return: generator of blobs
    """
    if mode == blobxfer.models.azure.StorageModes.File:
        raise RuntimeError('cannot list Azure Files from blob client')
    if blobxfer.util.blob_is_snapshot(prefix):
        base_blob, snapshot = blobxfer.util.parse_blob_snapshot_parameter(
            prefix)
        blob = client.get_blob_properties(
            container_name=container, blob_name=base_blob, snapshot=snapshot,
            timeout=timeout)
        yield blob
        return
    blobs = client.list_blobs(
        container_name=container,
        prefix=prefix if blobxfer.util.is_not_empty(prefix) else None,
        include=azure.storage.blob.models.Include.METADATA,
        timeout=timeout,
    )
    for blob in blobs:
        if (mode == blobxfer.models.azure.StorageModes.Append and
                blob.properties.blob_type !=
                azure.storage.blob.models._BlobTypes.AppendBlob):
            continue
        elif (mode == blobxfer.models.azure.StorageModes.Block and
                blob.properties.blob_type !=
                azure.storage.blob.models._BlobTypes.BlockBlob):
            continue
        elif (mode == blobxfer.models.azure.StorageModes.Page and
                blob.properties.blob_type !=
                azure.storage.blob.models._BlobTypes.PageBlob):
            continue
        if not recursive and '/' in blob.name:
            continue
        # auto or match, yield the blob
        yield blob


def list_all_blobs(client, container, timeout=None):
    # type: (azure.storage.blob.BaseBlobService, str, int) ->
    #        azure.storage.blob.models.Blob
    """List all blobs in a container
    :param azure.storage.blob.BaseBlobService client: blob client
    :param str container: container
    :param int timeout: timeout
    :rtype: azure.storage.blob.models.Blob
    :return: generator of blobs
    """
    blobs = client.list_blobs(
        container_name=container,
        prefix=None,
        timeout=timeout,
    )
    for blob in blobs:
        yield blob


def delete_blob(client, container, name, timeout=None):
    # type: (azure.storage.blob.BaseBlobService, str, str, int) -> None
    """Delete blob, including all associated snapshots
    :param azure.storage.blob.BaseBlobService client: blob client
    :param str container: container
    :param str name: blob name
    :param int timeout: timeout
    """
    client.delete_blob(
        container_name=container,
        blob_name=name,
        delete_snapshots=azure.storage.blob.models.DeleteSnapshot.Include,
        timeout=timeout,
    )  # noqa


def get_blob_range(ase, offsets, timeout=None):
    # type: (blobxfer.models.azure.StorageEntity,
    #        blobxfer.models.download.Offsets, int) -> bytes
    """Retrieve blob range
    :param blobxfer.models.azure.StorageEntity ase: Azure StorageEntity
    :param blobxfer.models.download.Offsets offsets: download offsets
    :param int timeout: timeout
    :rtype: bytes
    :return: content for blob range
    """
    return ase.client._get_blob(
        container_name=ase.container,
        blob_name=ase.name,
        snapshot=ase.snapshot,
        start_range=offsets.range_start,
        end_range=offsets.range_end,
        validate_content=False,  # HTTPS takes care of integrity during xfer
        timeout=timeout,
    ).content


def create_container(ase, containers_created, timeout=None):
    # type: (blobxfer.models.azure.StorageEntity, set, int) -> None
    """Create blob container
    :param blobxfer.models.azure.StorageEntity ase: Azure StorageEntity
    :param set containers_created: containers already created map
    :param int timeout: timeout
    """
    # check if auth allows create container
    if not ase.can_create_containers:
        return
    key = ase.client.account_name + ':blob=' + ase.container
    if key in containers_created:
        return
    if ase.client.create_container(
            container_name=ase.container,
            fail_on_exist=False,
            timeout=timeout):
        logger.info(
            'created blob container {} on storage account {}'.format(
                ase.container, ase.client.account_name))
    # always add to set (as it could be pre-existing)
    containers_created.add(key)


def set_blob_md5(ase, md5, timeout=None):
    # type: (blobxfer.models.azure.StorageEntity, str, int) -> None
    """Set blob properties MD5
    :param blobxfer.models.azure.StorageEntity ase: Azure StorageEntity
    :param str md5: md5 as base64
    :param int timeout: timeout
    """
    ase.client.set_blob_properties(
        container_name=ase.container,
        blob_name=ase.name,
        content_settings=azure.storage.blob.models.ContentSettings(
            content_type=blobxfer.util.get_mime_type(ase.name),
            content_md5=md5,
        ),
        timeout=timeout)  # noqa


def set_blob_metadata(ase, metadata, timeout=None):
    # type: (blobxfer.models.azure.StorageEntity, dict, int) -> None
    """Set blob metadata
    :param blobxfer.models.azure.StorageEntity ase: Azure StorageEntity
    :param dict metadata: metadata kv pairs
    :param int timeout: timeout
    """
    ase.client.set_blob_metadata(
        container_name=ase.container,
        blob_name=ase.name,
        metadata=metadata,
        timeout=timeout)  # noqa



# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import sys
from .version import __version__  # noqa

# monkeypatch User-Agent string
import azure.storage.common
azure.storage.common._constants.USER_AGENT_STRING_PREFIX = \
    'blobxfer/{} {}'.format(
        __version__,
        azure.storage.common._constants.USER_AGENT_STRING_PREFIX
    )

# set stdin source
if sys.version_info >= (3, 0):  # noqa
    STDIN = sys.stdin.buffer
else:  # noqa
    # set stdin to binary mode on Windows
    if sys.platform == 'win32':
        import msvcrt
        import os
        msvcrt.setmode(sys.stdin.fileno(), os.O_BINARY)
    STDIN = sys.stdin



# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# compat imports
from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
from builtins import (  # noqa
    bytes, dict, int, list, object, range, ascii, chr, hex, input,
    next, oct, open, pow, round, super, filter, map, zip)
# stdlib imports
import enum
try:
    import pathlib2 as pathlib
except ImportError:  # noqa
    import pathlib
# non-stdlib imports
from azure.storage.blob.models import _BlobTypes as BlobTypes
# local imports
import blobxfer.models.metadata


# enums
class StorageModes(enum.Enum):
    Auto = 10
    Append = 20
    Block = 30
    File = 40
    Page = 50


class StorageEntity(object):
    """Azure Storage Entity"""
    def __init__(self, container, ed=None):
        # type: (StorageEntity, str
        #        blobxfer.models.crypto.EncryptionMetadata) -> None
        """Ctor for StorageEntity
        :param StorageEntity self: this
        :param str container: container name
        :param blobxfer.models.crypto.EncryptionMetadata ed:
            encryption metadata
        """
        self._can_create_containers = None
        self._client = None
        self._container = container
        self._name = None
        self._mode = None
        self._lmt = None
        self._size = None
        self._snapshot = None
        self._md5 = None
        self._encryption = ed
        self._from_local = False
        self._append_create = True
        self._vio = None
        self._fileattr = None
        self._raw_metadata = None
        self._access_tier = None
        self.replica_targets = None

    @property
    def can_create_containers(self):
        # type: (StorageEntity) -> bool
        """Create containers
        :param StorageEntity self: this
        :rtype: bool
        :return: create containers
        """
        return self._can_create_containers

    @property
    def client(self):
        # type: (StorageEntity) -> object
        """Associated storage client
        :param StorageEntity self: this
        :rtype: object
        :return: associated storage client
        """
        return self._client

    @property
    def container(self):
        # type: (StorageEntity) -> str
        """Container name
        :param StorageEntity self: this
        :rtype: str
        :return: name of container or file share
        """
        return self._container

    @property
    def name(self):
        # type: (StorageEntity) -> str
        """Entity name
        :param StorageEntity self: this
        :rtype: str
        :return: name of entity
        """
        return self._name

    @property
    def path(self):
        # type: (StorageEntity) -> str
        """Entity path
        :param StorageEntity self: this
        :rtype: str
        :return: remote path of entity
        """
        return '{}/{}'.format(self._container, self._name)

    @property
    def lmt(self):
        # type: (StorageEntity) -> datetime.datetime
        """Entity last modified time
        :param StorageEntity self: this
        :rtype: datetime.datetime
        :return: LMT of entity
        """
        return self._lmt

    @property
    def size(self):
        # type: (StorageEntity) -> int
        """Entity size
        :param StorageEntity self: this
        :rtype: int
        :return: size of entity
        """
        return self._size

    @size.setter
    def size(self, value):
        # type: (StorageEntity, int) -> None
        """Set entity size
        :param StorageEntity self: this
        :param int value: value
        """
        self._size = value

    @property
    def snapshot(self):
        # type: (StorageEntity) -> str
        """Entity snapshot
        :param StorageEntity self: this
        :rtype: str
        :return: snapshot of entity
        """
        return self._snapshot

    @property
    def md5(self):
        # type: (StorageEntity) -> str
        """Base64-encoded MD5
        :param StorageEntity self: this
        :rtype: str
        :return: md5 of entity
        """
        return self._md5

    @property
    def mode(self):
        # type: (StorageEntity) -> blobxfer.models.azure.StorageModes
        """Entity mode (type)
        :param StorageEntity self: this
        :rtype: blobxfer.models.azure.StorageModes
        :return: type of entity
        """
        return self._mode

    @property
    def from_local(self):
        # type: (StorageEntity) -> bool
        """If entity was created from a local file (no remote exists)
        :param StorageEntity self: this
        :rtype: bool
        :return: if entity is from local (no remote exists)
        """
        return self._from_local

    @property
    def append_create(self):
        # type: (StorageEntity) -> bool
        """If append blob should be created
        :param StorageEntity self: this
        :rtype: bool
        :return: if append blob should be created
        """
        return self._append_create

    @append_create.setter
    def append_create(self, value):
        # type: (StorageEntity, bool) -> None
        """Set append create option
        :param StorageEntity self: this
        :param bool value: value to set
        """
        self._append_create = value

    @property
    def is_encrypted(self):
        # type: (StorageEntity) -> bool
        """If data is encrypted
        :param StorageEntity self: this
        :rtype: bool
        :return: if encryption metadata is present
        """
        return self._encryption is not None

    @property
    def encryption_metadata(self):
        # type: (StorageEntity) ->
        #        blobxfer.models.crypto.EncryptionMetadata
        """Get encryption metadata
        :param StorageEntity self: this
        :rtype: blobxfer.models.crypto.EncryptionMetadata
        :return: encryption metadata of entity
        """
        return self._encryption

    @encryption_metadata.setter
    def encryption_metadata(self, value):
        # type: (StorageEntity,
        #        blobxfer.models.crypto.EncryptionMetadata) -> None
        """Set encryption metadata
        :param StorageEntity self: this
        :param blobxfer.models.crypto.EncryptionMetadata value: value
        """
        self._encryption = value

    @property
    def file_attributes(self):
        # type: (StorageEntity) -> object
        """Return file attributes collection
        :param StorageEntity self: this
        :rtype: blobxfer.models.metadata.PosixFileAttr or
            blobxfer.models.metadata.WindowsFileAttr or None
        :return: file attributes
        """
        return self._fileattr

    @property
    def vectored_io(self):
        # type: (StorageEntity) -> object
        """Return vectored io metadata, currently stripe only
        :param StorageEntity self: this
        :rtype: blobxfer.models.metadata.VectoredStripe or None
        :return: vectored io metadata
        """
        return self._vio

    @property
    def raw_metadata(self):
        # type: (StorageEntity) -> dict
        """Return raw metadata for synccopy sources
        :param StorageEntity self: this
        :rtype: dict
        :return: raw metadata
        """
        return self._raw_metadata

    @property
    def access_tier(self):
        # type: (StorageEntity) -> str
        """Return access tier for blob
        :param StorageEntity self: this
        :rtype: str
        :return: access tier
        """
        return self._access_tier

    @access_tier.setter
    def access_tier(self, value):
        # type: (StorageEntity, str) -> None
        """Set access tier
        :param StorageEntity self: this
        :param str value: value
        """
        self._access_tier = value

    def populate_from_blob(self, sa, blob, vio=None, store_raw_metadata=False):
        # type: (StorageEntity, blobxfer.operations.azure.StorageAccount,
        #        azure.storage.blob.models.Blob) -> None
        """Populate properties from Blob
        :param StorageEntity self: this
        :param blobxfer.operations.azure.StorageAccount sa: storage account
        :param azure.storage.blob.models.Blob blob: blob to populate from
        :param blobxfer.models.metadata.VectoredStripe vio: Vectored stripe
        :param bool store_raw_metadata: store raw metadata
        """
        if store_raw_metadata:
            self._raw_metadata = blob.metadata
        else:
            self._fileattr = blobxfer.models.metadata.fileattr_from_metadata(
                blob.metadata)
        self._vio = vio
        self._can_create_containers = sa.can_create_containers
        self._name = blob.name
        self._snapshot = blob.snapshot
        self._lmt = blob.properties.last_modified
        self._size = blob.properties.content_length
        self._md5 = blob.properties.content_settings.content_md5
        if blob.properties.blob_type == BlobTypes.AppendBlob:
            self._mode = StorageModes.Append
            self._client = sa.append_blob_client
        elif blob.properties.blob_type == BlobTypes.BlockBlob:
            self._access_tier = blob.properties.blob_tier
            self._mode = StorageModes.Block
            self._client = sa.block_blob_client
        elif blob.properties.blob_type == BlobTypes.PageBlob:
            self._mode = StorageModes.Page
            self._client = sa.page_blob_client

    def populate_from_file(
            self, sa, file, path, vio=None, store_raw_metadata=False,
            snapshot=None):
        # type: (StorageEntity, blobxfer.operations.azure.StorageAccount,
        #        azure.storage.file.models.File, str,
        #        blobxfer.models.metadata.VectoredStripe, bool, str) -> None
        """Populate properties from File
        :param StorageEntity self: this
        :param blobxfer.operations.azure.StorageAccount sa: storage account
        :param azure.storage.file.models.File file: file to populate from
        :param str path: full path to file
        :param blobxfer.models.metadata.VectoredStripe vio: Vectored stripe
        :param bool store_raw_metadata: store raw metadata
        :param str snapshot: snapshot
        """
        if store_raw_metadata:
            self._raw_metadata = file.metadata
        else:
            self._fileattr = blobxfer.models.metadata.fileattr_from_metadata(
                file.metadata)
        self._vio = vio
        self._can_create_containers = sa.can_create_containers
        if path is not None:
            self._name = str(pathlib.Path(path) / file.name)
        else:
            self._name = file.name
        self._snapshot = snapshot
        self._lmt = file.properties.last_modified
        self._size = file.properties.content_length
        self._md5 = file.properties.content_settings.content_md5
        self._mode = StorageModes.File
        self._client = sa.file_client

    def populate_from_local(self, sa, container, path, mode):
        # type: (StorageEntity, blobxfer.operations.azure.StorageAccount
        #        str, str, blobxfer.models.azure.StorageModes) -> None
        """Populate properties from local
        :param StorageEntity self: this
        :param blobxfer.operations.azure.StorageAccount sa: storage account
        :param str container: container
        :param str path: full path to file
        :param blobxfer.models.azure.StorageModes mode: storage mode
        """
        self._can_create_containers = sa.can_create_containers
        self._container = container
        self._name = path
        self._mode = mode
        self._from_local = True
        if mode == StorageModes.Append:
            self._client = sa.append_blob_client
        elif mode == StorageModes.Block:
            self._client = sa.block_blob_client
        elif mode == StorageModes.File:
            self._client = sa.file_client
        elif mode == StorageModes.Page:
            self._client = sa.page_blob_client
        elif mode == StorageModes.Auto:
            name = self.name.lower()
            if name.endswith('.vhd') or name.endswith('.vhdx'):
                self._client = sa.page_blob_client
                self._mode = StorageModes.Page
            else:
                self._client = sa.block_blob_client
                self._mode = StorageModes.Block



# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# compat imports
from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
from builtins import (  # noqa
    bytes, dict, int, list, object, range, ascii, chr, hex, input,
    next, oct, open, pow, round, super, filter, map, zip)
# stdlib imports
import collections
import logging
import math
import threading
# non-stdlib imports
import bitstring
# local imports
import blobxfer.models.azure
import blobxfer.models.crypto
import blobxfer.models.options
import blobxfer.util

# create logger
logger = logging.getLogger(__name__)
# global defines
_MAX_NONBLOCK_BLOB_CHUNKSIZE_BYTES = 4194304
# named tuples
Offsets = collections.namedtuple(
    'Offsets', [
        'chunk_num',
        'num_bytes',
        'range_end',
        'range_start',
    ]
)


class Specification(object):
    """Synccopy Specification"""
    def __init__(self, synccopy_options, skip_on_options):
        # type: (Specification, blobxfer.models.options.SyncCopy,
        #        blobxfer.models.options.SkipOn, LocalDestinationPath) -> None
        """Ctor for Specification
        :param DownloadSpecification self: this
        :param blobxfer.models.options.SyncCopy synccopy_options:
            synccopy options
        :param blobxfer.models.options.SkipOn skip_on_options: skip on options
        """
        self.options = synccopy_options
        self.skip_on = skip_on_options
        self.sources = []
        self.destinations = []

    def add_azure_source_path(self, source):
        # type: (Specification, blobxfer.operations.azure.SourcePath) -> None
        """Add an Azure Source Path
        :param DownloadSpecification self: this
        :param blobxfer.operations.Azure.SourcePath source:
            Azure source path to add
        """
        self.sources.append(source)

    def add_azure_destination_path(self, dest):
        # type: (Specification,
        #        blobxfer.operations.azure.DestinationPath) -> None
        """Add a remote Azure Destination path
        :param UploadSpecification self: this
        :param blobxfer.operations.azure.DestinationPath dest:
            Remote destination path
        """
        self.destinations.append(dest)


class Descriptor(object):
    """Synccopy Descriptor"""
    def __init__(self, src_ase, dst_ase, block_list, options, resume_mgr):
        # type: (Descriptior, blobxfer.models.azure.StorageEntity,
        #        blobxfer.models.azure.StorageEntity, list,
        #        blobxfer.models.options.SyncCopy,
        #        blobxfer.operations.resume.SyncCopyResumeManager) -> None
        """Ctor for Descriptor
        :param Descriptor self: this
        :param blobxfer.models.azure.StorageEntity src_ase:
            source Azure Storage Entity
        :param blobxfer.models.azure.StorageEntity dst_ase:
            destination Azure Storage Entity
        :param list block_list: source blob block list
        :param blobxfer.models.options.SyncCopy options: synccopy options
        :param blobxfer.operations.resume.SyncCopyResumeManager resume_mgr:
            synccopy resume manager
        """
        self._offset = 0
        self._chunk_num = 0
        self._finalized = False
        self._meta_lock = threading.Lock()
        self._resume_mgr = resume_mgr
        self._src_ase = src_ase
        self._dst_ase = dst_ase
        self._src_block_list = block_list
        self._chunk_size = self._compute_chunk_size()
        # calculate the total number of ops required for transfer
        self._total_chunks = self._compute_total_chunks(self._chunk_size)
        self._outstanding_ops = self._total_chunks
        if blobxfer.util.is_not_empty(self._dst_ase.replica_targets):
            self._outstanding_ops *= len(self._dst_ase.replica_targets) + 1
        if self._resume_mgr:
            self._completed_chunks = bitstring.BitArray(
                length=self._total_chunks)
            self._replica_counters = {}

    @property
    def src_entity(self):
        # type: (Descriptor) -> blobxfer.models.azure.StorageEntity
        """Get linked source blobxfer.models.azure.StorageEntity
        :param Descriptor self: this
        :rtype: blobxfer.models.azure.StorageEntity
        :return: blobxfer.models.azure.StorageEntity
        """
        return self._src_ase

    @property
    def dst_entity(self):
        # type: (Descriptor) -> blobxfer.models.azure.StorageEntity
        """Get linked destination blobxfer.models.azure.StorageEntity
        :param Descriptor self: this
        :rtype: blobxfer.models.azure.StorageEntity
        :return: blobxfer.models.azure.StorageEntity
        """
        return self._dst_ase

    @property
    def all_operations_completed(self):
        # type: (Descriptor) -> bool
        """All operations are completed
        :param Descriptor self: this
        :rtype: bool
        :return: if all operations completed
        """
        with self._meta_lock:
            return self._outstanding_ops == 0

    @property
    def is_resumable(self):
        # type: (Descriptor) -> bool
        """Download is resume capable
        :param Descriptor self: this
        :rtype: bool
        :return: if resumable
        """
        return self._resume_mgr is not None

    @property
    def last_block_num(self):
        # type: (Descriptor) -> bool
        """Last used block number for block id, should only be called for
        finalize operation
        :param Descriptor self: this
        :rtype: int
        :return: block number
        """
        with self._meta_lock:
            return self._chunk_num - 1

    @property
    def remote_is_file(self):
        # type: (Descriptor) -> bool
        """Remote destination is an Azure File
        :param Descriptor self: this
        :rtype: bool
        :return: remote is an Azure File
        """
        return self.dst_entity.mode == blobxfer.models.azure.StorageModes.File

    @property
    def remote_is_page_blob(self):
        # type: (Descriptor) -> bool
        """Remote destination is an Azure Page Blob
        :param Descriptor self: this
        :rtype: bool
        :return: remote is an Azure Page Blob
        """
        return self.dst_entity.mode == blobxfer.models.azure.StorageModes.Page

    @property
    def remote_is_append_blob(self):
        # type: (Descriptor) -> bool
        """Remote destination is an Azure Append Blob
        :param Descriptor self: this
        :rtype: bool
        :return: remote is an Azure Append Blob
        """
        return (
            self.dst_entity.mode == blobxfer.models.azure.StorageModes.Append
        )

    @property
    def remote_is_block_blob(self):
        # type: (Descriptor) -> bool
        """Remote destination is an Azure Block Blob
        :param Descriptor self: this
        :rtype: bool
        :return: remote is an Azure Block Blob
        """
        return self.dst_entity.mode == blobxfer.models.azure.StorageModes.Block

    @property
    def is_one_shot_block_blob(self):
        # type: (Descriptor) -> bool
        """Is one shot block blob
        :param Descriptor self: this
        :rtype: bool
        :return: if upload is a one-shot block blob
        """
        return self.remote_is_block_blob and self._total_chunks == 1

    @property
    def requires_put_block_list(self):
        # type: (Descriptor) -> bool
        """Requires a put block list operation to finalize
        :param Descriptor self: this
        :rtype: bool
        :return: if finalize requires a put block list
        """
        return self.remote_is_block_blob and self._total_chunks > 1

    @property
    def requires_access_tier_set(self):
        # type: (Descriptor) -> bool
        """Remote destination requires an access tier set operation
        :param Descriptor self: this
        :rtype: bool
        :return: access tier is set
        """
        return (self.remote_is_block_blob and
                self.dst_entity.access_tier is not None)

    def complete_offset_upload(self, chunk_num):
        # type: (Descriptor, int) -> None
        """Complete the upload for the offset
        :param Descriptor self: this
        :param int chunk_num: chunk num completed
        """
        with self._meta_lock:
            self._outstanding_ops -= 1
            # save resume state
            if self.is_resumable:
                # only set resumable completed if all replicas for this
                # chunk are complete
                if blobxfer.util.is_not_empty(self._dst_ase.replica_targets):
                    if chunk_num not in self._replica_counters:
                        # start counter at -1 since we need 1 "extra" for the
                        # primary in addition to the replica targets
                        self._replica_counters[chunk_num] = -1
                    self._replica_counters[chunk_num] += 1
                    if (self._replica_counters[chunk_num] !=
                            len(self._dst_ase.replica_targets)):
                        return
                    else:
                        self._replica_counters.pop(chunk_num)
                self._completed_chunks.set(True, chunk_num)
                completed = self._outstanding_ops == 0
                self._resume_mgr.add_or_update_record(
                    self._dst_ase, self._src_block_list, self._offset,
                    self._chunk_size, self._total_chunks,
                    self._completed_chunks.int, completed,
                )

    def _compute_chunk_size(self):
        # type: (Descriptor) -> int
        """Compute chunk size given block list
        :param Descriptor self: this
        :rtype: int
        :return: chunk size bytes
        """
        if self._src_block_list is not None:
            blen = len(self._src_block_list)
            if blen == 0:
                # this is a one-shot block blob
                return self._src_ase.size
            elif blen == 1:
                return self._src_block_list[0].size
            else:
                return -1
        else:
            return _MAX_NONBLOCK_BLOB_CHUNKSIZE_BYTES

    def _compute_total_chunks(self, chunk_size):
        # type: (Descriptor, int) -> int
        """Compute total number of chunks for entity
        :param Descriptor self: this
        :param int chunk_size: chunk size
        :rtype: int
        :return: num chunks
        """
        try:
            if self._src_block_list is not None:
                blen = len(self._src_block_list)
                if blen > 0:
                    return blen
                else:
                    return 1
            else:
                return int(math.ceil(self._src_ase.size / chunk_size))
        except ZeroDivisionError:
            return 1

    def _resume(self):
        # type: (Descriptor) -> int
        """Resume a download, if possible
        :param Descriptor self: this
        :rtype: int or None
        :return: verified download offset
        """
        if self._resume_mgr is None or self._offset > 0:
            return None
        # check if path exists in resume db
        rr = self._resume_mgr.get_record(self._dst_ase)
        if rr is None:
            logger.debug('no resume record for {}'.format(self._dst_ase.path))
            return None
        # ensure lengths are the same
        if rr.length != self._src_ase.size:
            logger.warning('resume length mismatch {} -> {}'.format(
                rr.length, self._src_ase.size))
            return None
        # compute replica factor
        if blobxfer.util.is_not_empty(self._dst_ase.replica_targets):
            replica_factor = 1 + len(self._dst_ase.replica_targets)
        else:
            replica_factor = 1
        # set offsets if completed
        if rr.completed:
            with self._meta_lock:
                logger.debug('{} upload already completed'.format(
                    self._dst_ase.path))
                self._offset = rr.offset
                self._src_block_list = rr.src_block_list
                self._chunk_num = rr.total_chunks
                self._chunk_size = rr.chunk_size
                self._total_chunks = rr.total_chunks
                self._completed_chunks.int = rr.completed_chunks
                self._outstanding_ops = 0
                return self._src_ase.size * replica_factor
        # re-hash from 0 to offset if needed
        _cc = bitstring.BitArray(length=rr.total_chunks)
        _cc.int = rr.completed_chunks
        curr_chunk = _cc.find('0b0')[0]
        del _cc
        # set values from resume
        with self._meta_lock:
            self._offset = rr.offset
            self._src_block_list = rr.src_block_list
            self._chunk_num = curr_chunk
            self._chunk_size = rr.chunk_size
            self._total_chunks = rr.total_chunks
            self._completed_chunks = bitstring.BitArray(length=rr.total_chunks)
            self._completed_chunks.set(True, range(0, curr_chunk + 1))
            self._outstanding_ops = (
                (rr.total_chunks - curr_chunk) * replica_factor
            )
            logger.debug(
                ('resuming file {} from byte={} chunk={} chunk_size={} '
                 'total_chunks={} outstanding_ops={}').format(
                     self._src_ase.path, self._offset, self._chunk_num,
                     self._chunk_size, self._total_chunks,
                     self._outstanding_ops))
            return rr.offset * replica_factor

    def next_offsets(self):
        # type: (Descriptor) -> Offsets
        """Retrieve the next offsets
        :param Descriptor self: this
        :rtype: Offsets
        :return: download offsets
        """
        resume_bytes = self._resume()
        with self._meta_lock:
            if self._chunk_num >= self._total_chunks:
                return None, resume_bytes
            if self._chunk_size == -1 and self._src_block_list is not None:
                num_bytes = self._src_block_list[self._chunk_num].size
            else:
                if self._offset + self._chunk_size > self._src_ase.size:
                    num_bytes = self._src_ase.size - self._offset
                else:
                    num_bytes = self._chunk_size
            chunk_num = self._chunk_num
            range_start = self._offset
            range_end = self._offset + num_bytes - 1
            self._offset += num_bytes
            self._chunk_num += 1
            return Offsets(
                chunk_num=chunk_num,
                num_bytes=num_bytes,
                range_start=range_start,
                range_end=range_end,
            ), resume_bytes



# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# compat imports
from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
from builtins import (  # noqa
    bytes, dict, int, list, object, range, ascii, chr, hex, input,
    next, oct, open, pow, round, super, filter, map, zip)
# stdlib imports
import base64
import collections
import hashlib
import hmac
import json
import os
# non-stdlib imports
# local imports
import blobxfer.models.offload
import blobxfer.operations.crypto
import blobxfer.util

# encryption constants
AES256_BLOCKSIZE_BYTES = 16

# named tuples
EncryptionBlobxferExtensions = collections.namedtuple(
    'EncryptionBlobxferExtensions', [
        'pre_encrypted_content_md5',
    ]
)
EncryptionAgent = collections.namedtuple(
    'EncryptionAgent', [
        'encryption_algorithm',
        'protocol',
    ]
)
EncryptionAuthentication = collections.namedtuple(
    'EncryptionAuthentication', [
        'algorithm',
        'message_authentication_code',
    ]
)
EncryptionWrappedContentKey = collections.namedtuple(
    'EncryptionWrappedContentKey', [
        'algorithm',
        'encrypted_authentication_key',
        'encrypted_key',
        'key_id',
    ]
)
EncryptionMetadataAuthentication = collections.namedtuple(
    'EncryptionMetadataAuthentication', [
        'algorithm',
        'encoding',
        'message_authentication_code',
    ]
)


class EncryptionMetadata(object):
    """EncryptionMetadata"""

    # constants
    _ENCRYPTION_MODE = 'FullBlob'
    _ENCRYPTION_PROTOCOL_VERSION = '1.0'
    _ENCRYPTION_ALGORITHM = 'AES_CBC_256'
    _ENCRYPTED_KEY_SCHEME = 'RSA-OAEP'
    _AUTH_ALGORITHM = 'HMAC-SHA256'
    _AUTH_ENCODING_TYPE = 'UTF-8'

    _METADATA_KEY_NAME = 'encryptiondata'
    _METADATA_KEY_AUTH_NAME = 'encryptiondata_authentication'

    _JSON_KEY_ENCRYPTION_MODE = 'EncryptionMode'
    _JSON_KEY_ALGORITHM = 'Algorithm'
    _JSON_KEY_MAC = 'MessageAuthenticationCode'
    _JSON_KEY_ENCRYPTION_AGENT = 'EncryptionAgent'
    _JSON_KEY_PROTOCOL = 'Protocol'
    _JSON_KEY_ENCRYPTION_ALGORITHM = 'EncryptionAlgorithm'
    _JSON_KEY_INTEGRITY_AUTH = 'EncryptionAuthentication'
    _JSON_KEY_WRAPPEDCONTENTKEY = 'WrappedContentKey'
    _JSON_KEY_ENCRYPTED_KEY = 'EncryptedKey'
    _JSON_KEY_ENCRYPTED_AUTHKEY = 'EncryptedAuthenticationKey'
    _JSON_KEY_CONTENT_IV = 'ContentEncryptionIV'
    _JSON_KEY_KEYID = 'KeyId'
    _JSON_KEY_KEY_WRAPPING_METADATA = 'KeyWrappingMetadata'
    _JSON_KEY_BLOBXFER_EXTENSIONS = 'BlobxferExtensions'
    _JSON_KEY_PREENCRYPTED_MD5 = 'PreEncryptedContentMD5'

    _JSON_KEY_AUTH_METAAUTH = 'EncryptionMetadataAuthentication'
    _JSON_KEY_AUTH_ENCODING = 'Encoding'

    def __init__(self):
        # type: (EncryptionMetadata) -> None
        """Ctor for EncryptionMetadata
        :param EncryptionMetadata self: this
        """
        self.blobxfer_extensions = None
        self.content_encryption_iv = None
        self.encryption_agent = None
        self.encryption_authentication = None
        self.encryption_mode = None
        self.key_wrapping_metadata = {}
        self.wrapped_content_key = None
        self.encryption_metadata_authentication = None
        self._symkey = None
        self._signkey = None
        self._rsa_public_key = None

    @property
    def symmetric_key(self):
        # type: (EncryptionMetadata) -> bytes
        """Get symmetric key
        :param EncryptionMetadata self: this
        :rtype: bytes
        :return: symmetric key
        """
        return self._symkey

    @property
    def signing_key(self):
        # type: (EncryptionMetadata) -> bytes
        """Get singing key
        :param EncryptionMetadata self: this
        :rtype: bytes
        :return: signing key
        """
        return self._signkey

    @staticmethod
    def encryption_metadata_exists(md):
        # type: (dict) -> bool
        """Check if encryption metadata exists in json metadata
        :param dict md: metadata dictionary
        :rtype: bool
        :return: if encryption metadata exists
        """
        try:
            if blobxfer.util.is_not_empty(
                    md[EncryptionMetadata._METADATA_KEY_NAME]):
                return True
        except (KeyError, TypeError):
            pass
        return False

    def create_new_metadata(self, rsa_public_key):
        # type: (EncryptionMetadata,
        #        cryptography.hazmat.primitives.asymmetric.rsa.RSAPublicKey)
        #        -> None
        """Create new metadata entries for encryption (upload)
        :param EncryptionMetadata self: this
        :param cryptography.hazmat.primitives.asymmetric.rsa.RSAPublicKey:
            rsa public key
        """
        self._rsa_public_key = rsa_public_key
        self._symkey = os.urandom(
            blobxfer.operations.crypto._AES256_KEYLENGTH_BYTES)
        self._signkey = os.urandom(
            blobxfer.operations.crypto._AES256_KEYLENGTH_BYTES)
        self.content_encryption_iv = os.urandom(AES256_BLOCKSIZE_BYTES)
        self.encryption_agent = EncryptionAgent(
            encryption_algorithm=EncryptionMetadata._ENCRYPTION_ALGORITHM,
            protocol=EncryptionMetadata._ENCRYPTION_PROTOCOL_VERSION,
        )
        self.encryption_mode = EncryptionMetadata._ENCRYPTION_MODE

    def convert_from_json(self, md, entityname, rsaprivatekey):
        # type: (EncryptionMetadata, dict, str,
        #        cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateKey)
        #        -> None
        """Read metadata json into objects
        :param EncryptionMetadata self: this
        :param dict md: metadata dictionary
        :param str entityname: entity name
        :param rsaprivatekey: RSA private key
        :type rsaprivatekey:
            cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateKey
        """
        # populate from encryption data
        ed = json.loads(md[EncryptionMetadata._METADATA_KEY_NAME])
        try:
            self.blobxfer_extensions = EncryptionBlobxferExtensions(
                pre_encrypted_content_md5=ed[
                    EncryptionMetadata._JSON_KEY_BLOBXFER_EXTENSIONS][
                        EncryptionMetadata._JSON_KEY_PREENCRYPTED_MD5],
            )
        except KeyError:
            pass
        self.content_encryption_iv = base64.b64decode(
            ed[EncryptionMetadata._JSON_KEY_CONTENT_IV])
        self.encryption_agent = EncryptionAgent(
            encryption_algorithm=ed[
                EncryptionMetadata._JSON_KEY_ENCRYPTION_AGENT][
                    EncryptionMetadata._JSON_KEY_ENCRYPTION_ALGORITHM],
            protocol=ed[
                EncryptionMetadata._JSON_KEY_ENCRYPTION_AGENT][
                    EncryptionMetadata._JSON_KEY_PROTOCOL],
        )
        if (self.encryption_agent.encryption_algorithm !=
                EncryptionMetadata._ENCRYPTION_ALGORITHM):
            raise RuntimeError('{}: unknown block cipher: {}'.format(
                entityname, self.encryption_agent.encryption_algorithm))
        if (self.encryption_agent.protocol !=
                EncryptionMetadata._ENCRYPTION_PROTOCOL_VERSION):
            raise RuntimeError('{}: unknown encryption protocol: {}'.format(
                entityname, self.encryption_agent.protocol))
        self.encryption_authentication = EncryptionAuthentication(
            algorithm=ed[
                EncryptionMetadata._JSON_KEY_INTEGRITY_AUTH][
                    EncryptionMetadata._JSON_KEY_ALGORITHM],
            message_authentication_code=ed[
                EncryptionMetadata._JSON_KEY_INTEGRITY_AUTH][
                    EncryptionMetadata._JSON_KEY_MAC],
        )
        if (self.encryption_authentication.algorithm !=
                EncryptionMetadata._AUTH_ALGORITHM):
            raise RuntimeError(
                '{}: unknown integrity/auth method: {}'.format(
                    entityname, self.encryption_authentication.algorithm))
        self.encryption_mode = ed[
            EncryptionMetadata._JSON_KEY_ENCRYPTION_MODE]
        if self.encryption_mode != EncryptionMetadata._ENCRYPTION_MODE:
            raise RuntimeError(
                '{}: unknown encryption mode: {}'.format(
                    entityname, self.encryption_mode))
        try:
            _eak = ed[EncryptionMetadata._JSON_KEY_WRAPPEDCONTENTKEY][
                EncryptionMetadata._JSON_KEY_ENCRYPTED_AUTHKEY]
        except KeyError:
            _eak = None
        self.wrapped_content_key = EncryptionWrappedContentKey(
            algorithm=ed[
                EncryptionMetadata._JSON_KEY_WRAPPEDCONTENTKEY][
                    EncryptionMetadata._JSON_KEY_ALGORITHM],
            encrypted_authentication_key=_eak,
            encrypted_key=ed[
                EncryptionMetadata._JSON_KEY_WRAPPEDCONTENTKEY][
                    EncryptionMetadata._JSON_KEY_ENCRYPTED_KEY],
            key_id=ed[
                EncryptionMetadata._JSON_KEY_WRAPPEDCONTENTKEY][
                    EncryptionMetadata._JSON_KEY_KEYID],
        )
        if (self.wrapped_content_key.algorithm !=
                EncryptionMetadata._ENCRYPTED_KEY_SCHEME):
            raise RuntimeError('{}: unknown key encryption scheme: {}'.format(
                entityname, self.wrapped_content_key.algorithm))
        # if RSA key is a public key, stop here as keys cannot be decrypted
        if rsaprivatekey is None:
            return
        # decrypt symmetric key
        self._symkey = blobxfer.operations.crypto.\
            rsa_decrypt_base64_encoded_key(
                rsaprivatekey, self.wrapped_content_key.encrypted_key)
        # decrypt signing key, if it exists
        if blobxfer.util.is_not_empty(
                self.wrapped_content_key.encrypted_authentication_key):
            self._signkey = blobxfer.operations.crypto.\
                rsa_decrypt_base64_encoded_key(
                    rsaprivatekey,
                    self.wrapped_content_key.encrypted_authentication_key)
        else:
            self._signkey = None
        # populate from encryption data authentication
        try:
            eda = json.loads(md[EncryptionMetadata._METADATA_KEY_AUTH_NAME])
        except KeyError:
            pass
        else:
            self.encryption_metadata_authentication = \
                EncryptionMetadataAuthentication(
                    algorithm=eda[
                        EncryptionMetadata._JSON_KEY_AUTH_METAAUTH][
                            EncryptionMetadata._JSON_KEY_ALGORITHM],
                    encoding=eda[
                        EncryptionMetadata._JSON_KEY_AUTH_METAAUTH][
                            EncryptionMetadata._JSON_KEY_AUTH_ENCODING],
                    message_authentication_code=eda[
                        EncryptionMetadata._JSON_KEY_AUTH_METAAUTH][
                            EncryptionMetadata._JSON_KEY_MAC],
                )
            if (self.encryption_metadata_authentication.algorithm !=
                    EncryptionMetadata._AUTH_ALGORITHM):
                raise RuntimeError(
                    '{}: unknown integrity/auth method: {}'.format(
                        entityname,
                        self.encryption_metadata_authentication.algorithm))
            # verify hmac
            authhmac = base64.b64decode(
                self.encryption_metadata_authentication.
                message_authentication_code)
            bmeta = md[EncryptionMetadata._METADATA_KEY_NAME].encode(
                self.encryption_metadata_authentication.encoding)
            hmacsha256 = hmac.new(self._signkey, digestmod=hashlib.sha256)
            hmacsha256.update(bmeta)
            if hmacsha256.digest() != authhmac:
                raise RuntimeError(
                    '{}: encryption metadata authentication failed'.format(
                        entityname))

    def convert_to_json_with_mac(self, md5digest, hmacdigest):
        # type: (EncryptionMetadata, str, str) -> dict
        """Constructs metadata for encryption
        :param EncryptionMetadata self: this
        :param str md5digest: md5 digest
        :param str hmacdigest: hmac-sha256 digest (data)
        :rtype: dict
        :return: encryption metadata
        """
        # encrypt keys
        enc_content_key = blobxfer.operations.crypto.\
            rsa_encrypt_key_base64_encoded(
                None, self._rsa_public_key, self.symmetric_key)
        enc_sign_key = blobxfer.operations.crypto.\
            rsa_encrypt_key_base64_encoded(
                None, self._rsa_public_key, self.signing_key)
        # generate json
        encjson = {
            EncryptionMetadata._JSON_KEY_ENCRYPTION_MODE:
            EncryptionMetadata._ENCRYPTION_MODE,
            EncryptionMetadata._JSON_KEY_CONTENT_IV:
            blobxfer.util.base64_encode_as_string(self.content_encryption_iv),
            EncryptionMetadata._JSON_KEY_WRAPPEDCONTENTKEY: {
                EncryptionMetadata._JSON_KEY_KEYID: 'private:pem',
                EncryptionMetadata._JSON_KEY_ENCRYPTED_KEY: enc_content_key,
                EncryptionMetadata._JSON_KEY_ENCRYPTED_AUTHKEY: enc_sign_key,
                EncryptionMetadata._JSON_KEY_ALGORITHM:
                EncryptionMetadata._ENCRYPTED_KEY_SCHEME,
            },
            EncryptionMetadata._JSON_KEY_ENCRYPTION_AGENT: {
                EncryptionMetadata._JSON_KEY_PROTOCOL:
                EncryptionMetadata._ENCRYPTION_PROTOCOL_VERSION,
                EncryptionMetadata._JSON_KEY_ENCRYPTION_ALGORITHM:
                EncryptionMetadata._ENCRYPTION_ALGORITHM,
            },
            EncryptionMetadata._JSON_KEY_INTEGRITY_AUTH: {
                EncryptionMetadata._JSON_KEY_ALGORITHM:
                EncryptionMetadata._AUTH_ALGORITHM,
            },
            EncryptionMetadata._JSON_KEY_KEY_WRAPPING_METADATA: {},
        }
        if md5digest is not None:
            encjson[EncryptionMetadata._JSON_KEY_BLOBXFER_EXTENSIONS] = {
                EncryptionMetadata._JSON_KEY_PREENCRYPTED_MD5: md5digest
            }
        if hmacdigest is not None:
            encjson[EncryptionMetadata._JSON_KEY_INTEGRITY_AUTH][
                EncryptionMetadata._JSON_KEY_MAC] = hmacdigest
        bencjson = json.dumps(
            encjson, sort_keys=True, ensure_ascii=False).encode(
                EncryptionMetadata._AUTH_ENCODING_TYPE)
        encjson = {
            EncryptionMetadata._METADATA_KEY_NAME:
            json.dumps(encjson, sort_keys=True)
        }
        # compute MAC over encjson
        hmacsha256 = hmac.new(self._signkey, digestmod=hashlib.sha256)
        hmacsha256.update(bencjson)
        authjson = {
            EncryptionMetadata._JSON_KEY_AUTH_METAAUTH: {
                EncryptionMetadata._JSON_KEY_ALGORITHM:
                EncryptionMetadata._AUTH_ALGORITHM,
                EncryptionMetadata._JSON_KEY_AUTH_ENCODING:
                EncryptionMetadata._AUTH_ENCODING_TYPE,
                EncryptionMetadata._JSON_KEY_MAC:
                blobxfer.util.base64_encode_as_string(hmacsha256.digest()),
            }
        }
        encjson[EncryptionMetadata._METADATA_KEY_AUTH_NAME] = json.dumps(
            authjson, sort_keys=True)
        return encjson

    def initialize_hmac(self):
        # type: (EncryptionMetadata) -> hmac.HMAC
        """Initialize an hmac from a signing key if it exists
        :param EncryptionMetadata self: this
        :rtype: hmac.HMAC or None
        :return: hmac
        """
        if self._signkey is not None:
            return hmac.new(self._signkey, digestmod=hashlib.sha256)
        else:
            return None



from codecs import open
import re
try:
    from setuptools import setup
except ImportError:  # noqa
    from distutils.core import setup
import sys

if 'sdist' in sys.argv or 'bdist_wheel' in sys.argv:
    long_description = open('README.md', 'r', 'utf-8').read()
else:
    long_description = ''

with open('blobxfer/version.py', 'r', 'utf-8') as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
        fd.read(), re.MULTILINE).group(1)

if not version or len(version) == 0:
    raise RuntimeError('Cannot find version')

packages = [
    'blobxfer',
    'blobxfer.models',
    'blobxfer.operations',
    'blobxfer.operations.azure',
    'blobxfer.operations.azure.blob',
    'blobxfer_cli',
]

install_requires = [
    'azure-storage-blob>=1.3.1,<1.4.0',
    'azure-storage-file>=1.3.1,<1.4.0',
    'bitstring>=3.1.5,<3.2.0',
    'click>=7.0,<8.0',
    'cryptography>=2.3.1,<2.4.0',
    'future>=0.17.1<0.18.0',
    'pathlib2>=2.3.2;python_version<"3.5"',
    'python-dateutil>=2.7.5,<2.8.0',
    'requests>=2.20.0,<2.21.0',
    'ruamel.yaml>=0.15.76,<0.16.0',
    'scandir>=1.9.0;python_version<"3.5"',
]

setup(
    name='blobxfer',
    version=version,
    author='Microsoft Corporation',
    author_email='',
    description='Azure storage transfer tool and data movement library',
    platforms='any',
    url='https://github.com/Azure/blobxfer',
    license='MIT',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=packages,
    package_data={'blobxfer': ['LICENSE']},
    package_dir={'blobxfer': 'blobxfer', 'blobxfer_cli': 'cli'},
    entry_points={
        'console_scripts': 'blobxfer=blobxfer_cli.cli:cli',
    },
    zip_safe=False,
    install_requires=install_requires,
    tests_require=['pytest'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
    ],
    keywords=[
        'azure', 'storage', 'blob', 'files', 'transfer', 'copy', 'smb',
        'cifs', 'blobxfer', 'azcopy'
    ],
)



# Copyright (c) Microsoft Corporation
#
# All rights reserved.
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# compat imports
from __future__ import (
    absolute_import, division, print_function, unicode_literals
)
from builtins import (  # noqa
    bytes, dict, int, list, object, range, ascii, chr, hex, input,
    next, oct, open, pow, round, super, filter, map, zip)
# stdlib imports
import contextlib
import logging
try:
    import pathlib2 as pathlib
except ImportError:  # noqa
    import pathlib
import pickle
import shelve
import threading
# non-stdlib imports
# local imports
import blobxfer.models.resume
import blobxfer.util

# create logger
logger = logging.getLogger(__name__)


class _BaseResumeManager(object):
    """Base Resume Manager"""
    def __init__(self, resume_file):
        # type: (_BaseResumeManager, str) -> None
        """Ctor for _BaseResumeManager
        :param _BaseResumeManager self: this
        :param pathlib.Path resume_file: resume file
        """
        self._lock = threading.Lock()
        self._resume_file = resume_file
        self._data = shelve.open(
            str(resume_file), protocol=pickle.HIGHEST_PROTOCOL)

    def close(self):
        # type: (_BaseResumeManager) -> None
        """Close the internal data store
        :param _BaseResumeManager self: this
        """
        if self._data is not None:
            self._data.close()
            self._data = None

    def delete(self):
        # type: (_BaseResumeManager) -> None
        """Delete the resume file db
        :param _BaseResumeManager self: this
        """
        self.close()
        if self._resume_file.exists():  # noqa
            try:
                self._resume_file.unlink()
            except OSError as e:
                logger.warning('could not unlink resume db: {}'.format(e))
        for ext in ('.bak', '.dat', '.dir'):  # noqa
            fp = pathlib.Path(str(self._resume_file) + ext)
            if fp.exists():
                try:
                    fp.unlink()
                except OSError as e:
                    logger.warning('could not unlink resume db: {}'.format(e))

    @contextlib.contextmanager
    def datalock(self, acquire=True):
        # type: (_BaseResumeManager) -> None
        """Delete the resume file db
        :param _BaseResumeManager self: this
        :param bool acquire: acquire lock
        """
        if acquire:
            self._lock.acquire()
        try:
            yield
        finally:
            if acquire:
                self._lock.release()

    @staticmethod
    def generate_record_key(ase):
        # type: (blobxfer.models.azure.StorageEntity) -> str
        """Generate a record key
        :param blobxfer.models.azure.StorageEntity ase: Storage Entity
        :rtype: str
        :return: record key
        """
        key = '{}:{}'.format(ase._client.primary_endpoint, ase.path)
        if blobxfer.util.on_python2():
            return key.encode('utf8')
        else:
            return key

    def get_record(self, ase, key=None, lock=True):
        # type: (_BaseResumeManager, str, bool) -> object
        """Get a resume record
        :param _BaseResumeManager self: this
        :param blobxfer.models.azure.StorageEntity ase: Storage Entity
        :param str key: record key
        :param bool lock: acquire lock
        :rtype: object
        :return: resume record object
        """
        if key is None:
            key = blobxfer.operations.resume._BaseResumeManager.\
                generate_record_key(ase)
        with self.datalock(lock):
            try:
                return self._data[key]
            except KeyError:
                return None


class DownloadResumeManager(_BaseResumeManager):
    """Download Resume Manager"""
    def __init__(self, resume_file):
        # type: (DownloadResumeManager, str) -> None
        """Ctor for DownloadResumeManager
        :param DownloadResumeManager self: this
        :param pathlib.Path resume_file: resume file
        """
        super().__init__(resume_file)

    def add_or_update_record(
            self, final_path, ase, chunk_size, next_integrity_chunk,
            completed, md5):
        # type: (DownloadResumeManager, pathlib.Path,
        #        blobxfer.models.azure.StorageEntity, int, int, bool,
        #        str) -> None
        """Add or update a resume record
        :param DownloadResumeManager self: this
        :param pathlib.Path final_path: final path
        :param blobxfer.models.azure.StorageEntity ase: Storage Entity
        :param int chunk_size: chunk size in bytes
        :param int next_integrity_chunk: next integrity chunk
        :param bool completed: if completed
        :param str md5: md5 hex digest
        """
        key = blobxfer.operations.resume._BaseResumeManager.\
            generate_record_key(ase)
        with self.datalock():
            dl = self.get_record(ase, key=key, lock=False)
            if dl is None:
                dl = blobxfer.models.resume.Download(
                    final_path=str(final_path),
                    length=ase._size,
                    chunk_size=chunk_size,
                    next_integrity_chunk=next_integrity_chunk,
                    completed=completed,
                    md5=md5,
                )
            else:
                if (dl.completed or
                        next_integrity_chunk < dl.next_integrity_chunk):
                    return
                if completed:
                    dl.completed = completed
                else:
                    dl.next_integrity_chunk = next_integrity_chunk
                    dl.md5hexdigest = md5
            self._data[key] = dl
            self._data.sync()


class UploadResumeManager(_BaseResumeManager):
    """Upload Resume Manager"""
    def __init__(self, resume_file):
        # type: (UploadResumeManager, str) -> None
        """Ctor for UploadResumeManager
        :param UploadResumeManager self: this
        :param pathlib.Path resume_file: resume file
        """
        super().__init__(resume_file)

    def add_or_update_record(
            self, local_path, ase, chunk_size, total_chunks, completed_chunks,
            completed, md5):
        # type: (UploadResumeManager, pathlib.Path,
        #        blobxfer.models.azure.StorageEntity, int, int, int, bool,
        #        str) -> None
        """Add or update a resume record
        :param UploadResumeManager self: this
        :param pathlib.Path local_path: local path
        :param blobxfer.models.azure.StorageEntity ase: Storage Entity
        :param int chunk_size: chunk size in bytes
        :param int total_chunks: total chunks
        :param int completed_chunks: completed chunks bitarray
        :param bool completed: if completed
        :param str md5: md5 hex digest
        """
        key = blobxfer.operations.resume._BaseResumeManager.\
            generate_record_key(ase)
        with self.datalock():
            ul = self.get_record(ase, key=key, lock=False)
            if ul is None:
                ul = blobxfer.models.resume.Upload(
                    local_path=str(local_path),
                    length=ase._size,
                    chunk_size=chunk_size,
                    total_chunks=total_chunks,
                    completed_chunks=completed_chunks,
                    completed=completed,
                    md5=md5,
                )
            else:
                if ul.completed or completed_chunks == ul.completed_chunks:
                    return
                ul.completed_chunks = completed_chunks
                if completed:
                    ul.completed = completed
                else:
                    ul.md5hexdigest = md5
            self._data[key] = ul
            self._data.sync()


class SyncCopyResumeManager(_BaseResumeManager):
    """SyncCopy Resume Manager"""
    def __init__(self, resume_file):
        # type: (SyncCopyResumeManager, str) -> None
        """Ctor for SyncCopyResumeManager
        :param SyncCopyResumeManager self: this
        :param pathlib.Path resume_file: resume file
        """
        super().__init__(resume_file)

    def add_or_update_record(
            self, dst_ase, src_block_list, offset, chunk_size, total_chunks,
            completed_chunks, completed):
        # type: (SyncCopyResumeManager,
        #        blobxfer.models.azure.StorageEntity, list, int, int, int,
        #        int, bool) -> None
        """Add or update a resume record
        :param SyncCopyResumeManager self: this
        :param blobxfer.models.azure.StorageEntity dst_ase: Storage Entity
        :param list src_block_list: source block list
        :param int offset: offset
        :param int chunk_size: chunk size in bytes
        :param int total_chunks: total chunks
        :param int completed_chunks: completed chunks bitarray
        :param bool completed: if completed
        """
        key = blobxfer.operations.resume._BaseResumeManager.\
            generate_record_key(dst_ase)
        with self.datalock():
            sc = self.get_record(dst_ase, key=key, lock=False)
            if sc is None:
                sc = blobxfer.models.resume.SyncCopy(
                    length=dst_ase._size,
                    src_block_list=src_block_list,
                    offset=offset,
                    chunk_size=chunk_size,
                    total_chunks=total_chunks,
                    completed_chunks=completed_chunks,
                    completed=completed,
                )
            else:
                if sc.completed or completed_chunks == sc.completed_chunks:
                    return
                sc.offset = offset
                sc.completed_chunks = completed_chunks
                if completed:
                    sc.completed = completed
            self._data[key] = sc
            self._data.sync()
